import struct, abc, enum, typing
import iced_x86 as ix86

with open("FATTURA_.exe", "rb") as f:
    pe = f.read()

def read_u64(x: int) -> int:
    return struct.unpack("<Q", pe[x:x+8])[0]

def read_u32(x: int) -> int:
    return struct.unpack("<I", pe[x:x+4])[0]

def read_u16(x: int) -> int:
    return struct.unpack("<H", pe[x:x+2])[0]

def at_va(va: int) -> slice:
    rva = va - image_base
    for s in sections.values():
        if rva >= s["rva"] and rva < s["rva"] + s["va_size"]:
            return slice(rva - s["rva"] + s["raw_offset"], s["raw_size"] + s["raw_offset"])
    raise ValueError("RVA is out of range")

def va_to_offset(va: int) -> int:
    rva = va - image_base
    for s in sections.values():
        if rva >= s["rva"] and rva < s["rva"] + s["va_size"]:
            return rva - s["rva"] + s["raw_offset"]
    raise ValueError("RVA is out of range")


class BasicBlockEnd(enum.Enum):
    CONDITIONAL = 1,
    UNCONDITIONAL = 2,
    INDIRECT = 3,
    SPLIT = 2,
    NOT_DONE = 4

class BasicBlock(object):
    def __init__(self, start_va: int, instructions: typing.List[ix86.Instruction], machine_code: typing.List[bytes]):
        self._start_va = start_va
        
        self._next = None 
        self._branch = None

        self._split = None
        self._simplified = None 
        self._label = 0

        self._set_instructions(instructions, machine_code)

        
    def _set_instructions(self, instructions: typing.List[ix86.Instruction], machine_code: typing.List[bytes]):
        self._instructions = instructions
        self._machine_code = machine_code
        self._update_end()
        self._update_size()

    
    def set_next(self, bb: "BasicBlock"):
        self._next = bb
    
    def set_branch(self, bb: "BasicBlock"):
        self._branch = bb

    def set_simplified(self, instructions: typing.List[ix86.Instruction]):
        self._simplified = instructions 

    @property 
    def simplified(self) -> typing.List[ix86.Instruction]:
        return self._simplified

    def _update_size(self):
        self._size = sum(map(lambda i: i.len, self._instructions)) 

    def _update_end(self):
        if not self._instructions:
            self._end = BasicBlockEnd.NOT_DONE
        elif self._instructions[-1].flow_control in [ix86.FlowControl.CONDITIONAL_BRANCH]:
            self._end = BasicBlockEnd.CONDITIONAL 
        elif self._instructions[-1].flow_control in [ix86.FlowControl.UNCONDITIONAL_BRANCH]:
            self._end = BasicBlockEnd.UNCONDITIONAL 
        elif self._instructions[-1].flow_control in [ix86.FlowControl.INDIRECT_BRANCH, ix86.FlowControl.RETURN]:
            self._end = BasicBlockEnd.INDIRECT  
        else:
            self._end = BasicBlockEnd.NOT_DONE

    def _add_instruction(self, instruction: ix86.Instruction, opcodes: bytes):
        self._instructions.append(instruction)
        self._machine_code.append(opcodes)
        self._update_end()
        self._update_size()   

    @property 
    def machine_code(self) -> typing.List[bytes]:
        return self._machine_code

    @property 
    def size(self) -> int:
        return self._size 

    @property 
    def instructions(self) -> typing.List[ix86.Instruction]:
        return self._instructions

    @property 
    def va(self) -> int:
        return self._start_va 

    @property 
    def end(self) -> BasicBlockEnd:
        return self._end 
    
    @property
    def next(self) -> "BasicBlock":
        return self._next 
    
    @property
    def branch(self) -> "BasicBlock":
        return self._branch 

    def encompass(self, va: int) -> bool:
        return va >= self.va and va < self.va + self.size

    def set_label(self, id: int):
        self._label = id 
    
    @property
    def label(self) -> int:
        return self._label

    #Returns the simplified instructions without any superflous instructions or the original instruction
    #In any case an unconditional trailing jump is removed and a conditional trailing jump will be added
    #if needed
    def new_instructions(self, ticket_off:int) -> typing.List[ix86.Instruction]:
        cpu = CPU()
        if self.simplified is not None:
            ins = [i for i in self.simplified if not cpu.is_superflous(i, ticket_off)]
        else:
            ins = self.instructions

        if self.branch is None and self.next is None:
            return ins 

        if self.branch is None and len(ins) > 0 and ins[-1].flow_control != ix86.FlowControl.NEXT:
            return ins[:-1]
        
        assert(self.branch is None or len(ins) > 0)
        if self.branch is not None and (len(ins) == 0 or ins[-1].flow_control != ix86.FlowControl.CONDITIONAL_BRANCH):
            ins.append(ix86.Instruction.create_branch(ix86.Code.JNE_REL32_64, 0))
        return ins

    def _begin_split(self, va: int) -> "BasicBlock":
        if not self.encompass(va):
            raise ValueError("VA is not inside this basic block")

        next_block = [(i, c) for i, c in zip(self._instructions, self._machine_code) if i.ip >= va]
        this_block = [(i, c) for i, c in zip(self._instructions, self._machine_code) if i.ip < va]

        if next_block[0][0].ip != va:
            raise ValueError("VA splits an instruction")
        next_block = BasicBlock(va, [i for i,_ in next_block], [c for _,c in next_block])
        next_block._branch = self._branch
        next_block._next = self._next 

        self._split = dict(
            instructions = [i for i,_ in this_block],
            machine_code = [c for _,c in this_block],
            branch = None,
            next = next_block)

        return next_block

    def _finish_split(self):
        if self._split:
            self._split["next"]._branch = self._branch
            self._split["next"]._next = self._next
            self._set_instructions(self._split["instructions"], self._split["machine_code"])
            self._branch = self._split["branch"]
            self._next = self._split["next"]
            self._end = BasicBlockEnd.SPLIT
            
            self._split = None 

    def __str__(self) -> str:
        return f"{hex(self.va)}: {('next ' + hex(self.next.va)) if self.next else ''} {('branch ' + hex(self.branch.va)) if self.branch else ''} ({self.size} bytes)"

    def __repr__(self) -> str:
        return hex(self.va)

    @classmethod
    def walk_function(cls, va: int, context: typing.Dict[int, "BasicBlock"]) -> "BasicBlock":
        #print("Walk at: " + hex(va))
        if va in context:
            #print("BB already done")
            return context[va] 

        containing_bb = [bb for bb in context.values() if bb.encompass(va)]
        if containing_bb:
            #print(f"Splitting bb at {hex(containing_bb[0].va)}")
            splitted = containing_bb[0]._begin_split(va)
            context[va] = splitted
            return splitted

        bb = BasicBlock(va, [], [])
        context[va] = bb
        #print("Ready " + str(len(pe[at_va(va)])))
        code = pe[at_va(va)]
        offset = 0
        decoder = ix86.Decoder(64, pe[at_va(va)], ip=va)
        for i in decoder:
            bb._add_instruction(i, code[offset:offset+i.len])
            offset += i.len

            #print(bb.end)
            if bb.end == BasicBlockEnd.CONDITIONAL:
                next_bb = cls.walk_function(i.next_ip, context)
                if bb._next is None:
                    bb._next = next_bb
                bb._branch = cls.walk_function(i.near_branch_target, context)
                bb._finish_split()
                return bb 
            if bb.end == BasicBlockEnd.UNCONDITIONAL:
                bb._next = cls.walk_function(i.near_branch_target, context)
                bb._finish_split()
                return bb 
            if bb.end == BasicBlockEnd.INDIRECT:
                bb._finish_split()
                return bb 


class DataSize(object):
    def __init__(self, size: int, format: str):
        self._size = size 
        self._format = format 
    
    @property 
    def size(self):
        return self._size 

    @property 
    def format(self):
        return self._format     

SIZE_B = DataSize(1, "<B")
SIZE_W = DataSize(2, "<H")
SIZE_D = DataSize(4, "<I")
SIZE_Q = DataSize(8, "<Q")



def read_data(va: int, size: DataSize) -> int | None:
    rva = va - image_base
    for s in sections.values():
        if rva >= s["rva"] and rva + size.size - 1 < s["rva"] + s["va_size"]:
            if rva - s["rva"] < s["raw_size"]:
                start = rva - s["rva"] + s["raw_offset"]
                return struct.unpack(size.format, pe[start:start+size.size])[0]
            else:
                return 0
    return None 



class CPU(object):
    def __init__(self):
        self._info = ix86.InstructionInfoFactory()
        self._reg_hierarchy = (
            (ix86.Register.AL, ix86.Register.AX, ix86.Register.EAX, ix86.Register.RAX),
            (ix86.Register.AH, ix86.Register.AX, ix86.Register.EAX, ix86.Register.RAX),
            (ix86.Register.BL, ix86.Register.BX, ix86.Register.EBX, ix86.Register.RBX),
            (ix86.Register.BH, ix86.Register.BX, ix86.Register.EBX, ix86.Register.RBX),
            (ix86.Register.CL, ix86.Register.CX, ix86.Register.ECX, ix86.Register.RCX),
            (ix86.Register.CH, ix86.Register.CX, ix86.Register.ECX, ix86.Register.RCX),
            (ix86.Register.DL, ix86.Register.DX, ix86.Register.EDX, ix86.Register.RDX),
            (ix86.Register.DH, ix86.Register.DX, ix86.Register.EDX, ix86.Register.RDX),
            (ix86.Register.DIL, ix86.Register.DI, ix86.Register.EDI, ix86.Register.RDI),
            (ix86.Register.SIL, ix86.Register.SI, ix86.Register.ESI, ix86.Register.RSI),
            (ix86.Register.BPL, ix86.Register.BP, ix86.Register.EBP, ix86.Register.RBP),
            (ix86.Register.SPL, ix86.Register.SP, ix86.Register.ESP, ix86.Register.RSP),
            (ix86.Register.R8L, ix86.Register.R8W, ix86.Register.R8D, ix86.Register.R8),
            (ix86.Register.R9L, ix86.Register.R9W, ix86.Register.R9D, ix86.Register.R9),
            (ix86.Register.R10L, ix86.Register.R10W, ix86.Register.R10D, ix86.Register.R10),
            (ix86.Register.R11L, ix86.Register.R11W, ix86.Register.R11D, ix86.Register.R11),
            (ix86.Register.R12L, ix86.Register.R12W, ix86.Register.R12D, ix86.Register.R12),
            (ix86.Register.R13L, ix86.Register.R13W, ix86.Register.R13D, ix86.Register.R13),
            (ix86.Register.R14L, ix86.Register.R14W, ix86.Register.R14D, ix86.Register.R14),
            (ix86.Register.R15L, ix86.Register.R15W, ix86.Register.R15D, ix86.Register.R15),
        )
        self._arith_mnemonic = [ix86.Mnemonic.SUB, ix86.Mnemonic.ADD, ix86.Mnemonic.XOR, ix86.Mnemonic.OR, ix86.Mnemonic.AND]
        self._arith_ops = {ix86.Mnemonic.SUB: lambda x,y: x-y, ix86.Mnemonic.ADD: lambda x,y: x+y, ix86.Mnemonic.XOR: lambda x,y: x^y, ix86.Mnemonic.OR: lambda x,y: x|y, ix86.Mnemonic.AND: lambda x,y: x&y}
        
    def _is_reg_in_reg(self, x: ix86.Register, y: ix86.Register) -> bool:
        for hr in self._reg_hierarchy:
            try:
                idx = hr.index(x)
            except ValueError:
                continue 

            try:
                hr.index(y, idx)
            except ValueError:
                continue 
                
            return True 
        return False

    def _get_output_registers(self, x: ix86.Instruction) -> typing.List[int]:
        info_x = self._info.info(x)
        return [rx.register for rx in info_x.used_registers() if rx.access in [ix86.OpAccess.COND_WRITE, ix86.OpAccess.READ_COND_WRITE, ix86.OpAccess.READ_WRITE, ix86.OpAccess.WRITE]]

    def _get_input_registers(self, x: ix86.Instruction) -> typing.List[int]:
        info_x = self._info.info(x)
        return [rx.register for rx in info_x.used_registers() if rx.access in [ix86.OpAccess.COND_READ, ix86.OpAccess.READ_COND_WRITE, ix86.OpAccess.READ_WRITE, ix86.OpAccess.READ]]

    def _reg_sets_overlap(self, rxs, rys) -> bool:
        for ry in rys:
            for rx in rxs:
                if self._is_reg_in_reg(ry, rx):
                    return True 
        return False

    def _conflict(self, x: ix86.Instruction, y: ix86.Instruction) -> bool:
        if x.rflags_modified & y.rflags_modified:
            return True

        rxs = self._get_output_registers(x) 
        if self._reg_sets_overlap(rxs, self._get_output_registers(y)):
            return True 

        if self._reg_sets_overlap(rxs, self._get_input_registers(y)):
            return True 

        return False

    def _write_to(self, x: ix86.Instruction, reg: ix86.Register) -> bool:
        rxs = self._get_output_registers(x) 
        for rx in rxs:
                if self._is_reg_in_reg(rx, reg):
                    return True
        return False
        
    @staticmethod
    def _is_op_immediate(ins: ix86.Instruction, op: int) -> bool:
        try:
            ins.immediate(op)
            return True
        except ValueError:
            return False

    @classmethod
    def _is_mov_immediate(cls, ins: ix86.Instruction) -> bool:
        return ins.mnemonic == ix86.Mnemonic.MOV and cls._is_op_immediate(ins, 1) and ins.op0_kind == ix86.OpKind.REGISTER

    def _reg_size(self, reg: ix86.Register) -> int:
        for hr in self._reg_hierarchy:
            try:
                return 8<<hr.index(reg)
            except ValueError:
                pass
        raise ValueError(f"Unhandled register for size: {reg}")

    @staticmethod
    def _reg_shift(reg: ix86.Register) -> int:
        return 8 if reg in [ix86.Register.AH, ix86.Register.BH, ix86.Register.CH, ix86.Register.DH] else 0

    def _make_ops(self, op_mnemonic: ix86.Mnemonic, op_size: int, op_shift: int) -> typing.Tuple[typing.Callable[[int, int], int], typing.Callable[[int], int], typing.Callable[[int, int], int]]:
        op_size_mask = ((1<<op_size)-1)
        op_mask = op_size_mask << op_shift 
        op = self._arith_ops[op_mnemonic]
        extract = lambda x: (x & op_mask) >> op_shift 
        insert = lambda x,y: (x & ~op_mask) | ((y & op_size_mask) << op_shift)
        return op, extract, insert

    def _is_arith_reg_imm(self, ins: ix86.Instruction, reg: ix86.Register) -> typing.Callable[[int, int], int]|None:
        if not (
            ins.mnemonic in self._arith_mnemonic 
            and ins.op0_kind == ix86.OpKind.REGISTER and self._is_op_immediate(ins, 1) and self._is_reg_in_reg(ins.op0_register, reg)
        ):
            return None

        op_size = self._reg_size(ins.op0_register)
        op_shift = self._reg_shift(ins.op0_register)
        op, extract, insert = self._make_ops(ins.mnemonic, op_size, op_shift)
        return lambda x, y: insert(x, op(extract(x), y))

    def _is_arith_reg_reg(self, ins: ix86.Instruction, reg: ix86.Register) -> typing.Callable[[int], int] | None:
        if not (
            ins.mnemonic in (self._arith_mnemonic + [ix86.Mnemonic.MOV])
            and ins.op0_kind == ix86.OpKind.REGISTER and ins.op1_kind == ix86.OpKind.REGISTER and self._is_reg_in_reg(ins.op1_register, reg)
        ):
            return None

        op_size = self._reg_size(ins.op0_register)
        op_shift = self._reg_shift(ins.op0_register)
        _, extract, _ = self._make_ops(self._arith_mnemonic[0], op_size, op_shift)

        return extract
       
    def _update_mov(self, ins: ix86.Instruction, immediate: int)-> ix86.Instruction:
        rs = self._reg_size(ins.op0_register)

        if rs in [8, 16, 32]:
            return ix86.Instruction.create_reg_u32(ins.code, ins.op0_register, immediate)
        return ix86.Instruction.create_reg_u64(ins.code, ins.op0_register, immediate) 

    def _make_mov_imm(self, reg: ix86.Register, immediate: int)-> ix86.Instruction:
        rs = self._reg_size(reg)

        if rs == 8:
            code = ix86.Code.MOV_R8_IMM8
        elif rs == 16:
            code = ix86.Code.MOV_R16_IMM16
        elif rs == 32:
            code = ix86.Code.MOV_R32_IMM32
        else:
            code = ix86.Code.MOV_R64_IMM64
        
        return ix86.Instruction.create_reg_u32(code, reg, immediate)

    def _update_arith(self, ins: ix86.Instruction, immediate: int)-> ix86.Instruction:
        updates= {
            ix86.Code.MOV_R8_RM8: ix86.Code.MOV_R8_IMM8,
            ix86.Code.MOV_RM8_R8: ix86.Code.MOV_R8_IMM8,
            ix86.Code.MOV_R16_RM16: ix86.Code.MOV_R16_IMM16,
            ix86.Code.MOV_RM16_R16: ix86.Code.MOV_R16_IMM16,
            ix86.Code.MOV_R32_RM32: ix86.Code.MOV_R32_IMM32,
            ix86.Code.MOV_RM32_R32: ix86.Code.MOV_R32_IMM32,
            ix86.Code.MOV_R64_RM64: ix86.Code.MOV_R64_IMM64,
            ix86.Code.MOV_RM64_R64: ix86.Code.MOV_R64_IMM64,


            ix86.Code.ADD_AL_IMM8: ix86.Code.ADD_RM8_IMM8,
            ix86.Code.ADD_R8_RM8: ix86.Code.ADD_RM8_IMM8,
            ix86.Code.ADD_RM8_R8: ix86.Code.ADD_RM8_IMM8,
            ix86.Code.ADD_AX_IMM16: ix86.Code.ADD_RM16_IMM16,
            ix86.Code.ADD_R16_RM16: ix86.Code.ADD_RM16_IMM16,
            ix86.Code.ADD_RM16_R16: ix86.Code.ADD_RM16_IMM16,
            ix86.Code.ADD_EAX_IMM32: ix86.Code.ADD_RM32_IMM32,
            ix86.Code.ADD_R32_RM32: ix86.Code.ADD_RM32_IMM32,
            ix86.Code.ADD_RM32_R32: ix86.Code.ADD_RM32_IMM32,
            ix86.Code.ADD_RAX_IMM32: ix86.Code.ADD_RM64_IMM32,
            ix86.Code.ADD_R64_RM64: ix86.Code.ADD_RM64_IMM32,
            ix86.Code.ADD_RM64_R64: ix86.Code.ADD_RM64_IMM32,

            ix86.Code.SUB_AL_IMM8:      ix86.Code.SUB_RM8_IMM8,
            ix86.Code.SUB_R8_RM8:       ix86.Code.SUB_RM8_IMM8,
            ix86.Code.SUB_RM8_R8:       ix86.Code.SUB_RM8_IMM8,
            ix86.Code.SUB_AX_IMM16:     ix86.Code.SUB_RM16_IMM16,
            ix86.Code.SUB_R16_RM16:     ix86.Code.SUB_RM16_IMM16,
            ix86.Code.SUB_RM16_R16:     ix86.Code.SUB_RM16_IMM16,
            ix86.Code.SUB_EAX_IMM32:    ix86.Code.SUB_RM32_IMM32,
            ix86.Code.SUB_R32_RM32:     ix86.Code.SUB_RM32_IMM32,
            ix86.Code.SUB_RM32_R32:     ix86.Code.SUB_RM32_IMM32,
            ix86.Code.SUB_RAX_IMM32:    ix86.Code.SUB_RM64_IMM32,
            ix86.Code.SUB_R64_RM64:     ix86.Code.SUB_RM64_IMM32,
            ix86.Code.SUB_RM64_R64:     ix86.Code.SUB_RM64_IMM32,

            ix86.Code.OR_AL_IMM8:      ix86.Code.OR_RM8_IMM8,
            ix86.Code.OR_R8_RM8:       ix86.Code.OR_RM8_IMM8,
            ix86.Code.OR_RM8_R8:       ix86.Code.OR_RM8_IMM8,
            ix86.Code.OR_AX_IMM16:     ix86.Code.OR_RM16_IMM16,
            ix86.Code.OR_R16_RM16:     ix86.Code.OR_RM16_IMM16,
            ix86.Code.OR_RM16_R16:     ix86.Code.OR_RM16_IMM16,
            ix86.Code.OR_EAX_IMM32:    ix86.Code.OR_RM32_IMM32,
            ix86.Code.OR_R32_RM32:     ix86.Code.OR_RM32_IMM32,
            ix86.Code.OR_RM32_R32:     ix86.Code.OR_RM32_IMM32,
            ix86.Code.OR_RAX_IMM32:    ix86.Code.OR_RM64_IMM32,
            ix86.Code.OR_R64_RM64:     ix86.Code.OR_RM64_IMM32,
            ix86.Code.OR_RM64_R64:     ix86.Code.OR_RM64_IMM32,

            ix86.Code.AND_AL_IMM8:      ix86.Code.AND_RM8_IMM8,
            ix86.Code.AND_R8_RM8:       ix86.Code.AND_RM8_IMM8,
            ix86.Code.AND_RM8_R8:       ix86.Code.AND_RM8_IMM8,
            ix86.Code.AND_AX_IMM16:     ix86.Code.AND_RM16_IMM16,
            ix86.Code.AND_R16_RM16:     ix86.Code.AND_RM16_IMM16,
            ix86.Code.AND_RM16_R16:     ix86.Code.AND_RM16_IMM16,
            ix86.Code.AND_EAX_IMM32:    ix86.Code.AND_RM32_IMM32,
            ix86.Code.AND_R32_RM32:     ix86.Code.AND_RM32_IMM32,
            ix86.Code.AND_RM32_R32:     ix86.Code.AND_RM32_IMM32,
            ix86.Code.AND_RAX_IMM32:    ix86.Code.AND_RM64_IMM32,
            ix86.Code.AND_R64_RM64:     ix86.Code.AND_RM64_IMM32,
            ix86.Code.AND_RM64_R64:     ix86.Code.AND_RM64_IMM32,

            ix86.Code.XOR_AL_IMM8:      ix86.Code.XOR_RM8_IMM8,
            ix86.Code.XOR_R8_RM8:       ix86.Code.XOR_RM8_IMM8,
            ix86.Code.XOR_RM8_R8:       ix86.Code.XOR_RM8_IMM8,
            ix86.Code.XOR_AX_IMM16:     ix86.Code.XOR_RM16_IMM16,
            ix86.Code.XOR_R16_RM16:     ix86.Code.XOR_RM16_IMM16,
            ix86.Code.XOR_RM16_R16:     ix86.Code.XOR_RM16_IMM16,
            ix86.Code.XOR_EAX_IMM32:    ix86.Code.XOR_RM32_IMM32,
            ix86.Code.XOR_R32_RM32:     ix86.Code.XOR_RM32_IMM32,
            ix86.Code.XOR_RM32_R32:     ix86.Code.XOR_RM32_IMM32,
            ix86.Code.XOR_RAX_IMM32:    ix86.Code.XOR_RM64_IMM32,
            ix86.Code.XOR_R64_RM64:     ix86.Code.XOR_RM64_IMM32,
            ix86.Code.XOR_RM64_R64:     ix86.Code.XOR_RM64_IMM32,
        }

        rs = self._reg_size(ins.op0_register)

        if rs in [8, 16, 32]:
            return ix86.Instruction.create_reg_u32(updates[ins.code], ins.op0_register, immediate)
        return ix86.Instruction.create_reg_u64(updates[ins.code], ins.op0_register, immediate)

    def _is_mov_to_lvar(self, i: ix86.Instruction, offset: int) -> bool:
        return (i.code in [ix86.Code.MOV_RM32_IMM32] and 
                i.memory_base == ix86.Register.RBP and 
                i.memory_displacement == offset and 
                i.memory_index == ix86.Register.NONE)

    def find_mov_to_local_var(self, offset: int, instructions: typing.List[ix86.Instruction]) -> typing.Tuple[int|None, int|None]:
        """
mov     al, [rbp+350h+var_329]
test    al, 1
mov     ecx, 3074A3A9h
mov     edx, 2DAD4BE8h
cmovnz  edx, ecx
mov     [rbp+350h+pilot1], edx
jmp     loc_4100C3
        """
        if len(instructions) == 6 and (

            instructions[0].code == ix86.Code.MOV_R8_RM8 and instructions[0].memory_base == ix86.Register.RBP and 
            instructions[1].code == ix86.Code.TEST_AL_IMM8 and
            instructions[2].code == ix86.Code.MOV_R32_IMM32 and instructions[2].op0_kind == ix86.OpKind.REGISTER and 
            instructions[3].code == ix86.Code.MOV_R32_IMM32 and instructions[3].op0_kind == ix86.OpKind.REGISTER and
            instructions[4].code == ix86.Code.CMOVNE_R32_RM32
        ):
            return instructions[3].immediate(1), instructions[2].immediate(1)
        
        for i in instructions:
            if (
                i.code not in [ix86.Code.MOV_RM32_IMM32] or 
                i.memory_base != ix86.Register.RBP or 
                i.memory_displacement != offset or 
                i.memory_index != ix86.Register.NONE):
                continue 

            return i.immediate(1), None
        return None, None

    def process(self, bb: BasicBlock):
        instructions = bb.instructions[:-1]
        modified = True
        while modified:
            modified = False 
            modified |= self._rewrite_zero_idioms(instructions)
            modified |= self._rewrite_loads(instructions)
            modified |= self._fold_constants(instructions)
            modified |= self._remove_dead_movs(instructions)
            modified |= self._remove_imul(instructions)
            modified |= self._remove_cmp_setcc(instructions)
            modified |= self._merge_add_sub(instructions)
            modified |= self._remove_cmovcc(instructions)

        def to_remove(i: ix86.Instruction) -> bool:
            return i.mnemonic == ix86.Mnemonic.NOP

        instructions = [x for x in instructions if not to_remove(x)]
        #for i in instructions:
        #    print(i)
        return instructions
    
    def _rewrite_zero_idioms(self, instructions: typing.List[ix86.Instruction]):
        modified = False
        for i, ins in enumerate(instructions):
            if ins.mnemonic in [ix86.Mnemonic.XOR, ix86.Mnemonic.SUB] and ins.op0_kind == ix86.OpKind.REGISTER and ins.op1_kind == ix86.OpKind.REGISTER and ins.op0_register ==ins.op1_register:
                instructions[i] = ix86.Instruction.create_reg_u32( ix86.Code.MOV_R32_IMM32, ins.op0_register, 0)
                modified = True 
        return modified

    def _rewrite_loads(self, instructions: typing.List[ix86.Instruction]):
        modified = False
        for i, ins in enumerate(instructions):
            if ins.mnemonic == ix86.Mnemonic.MOV and ins.op0_kind == ix86.OpKind.REGISTER and ins.op1_kind == ix86.OpKind.MEMORY and (
                ins.memory_base in [ix86.Register.NONE, ix86.Register.RIP] and ins.memory_index == ix86.Register.NONE):
                data = read_data(ins.memory_displacement, SIZE_D)
                if data is not None:
                    instructions[i] = ix86.Instruction.create_reg_u32( ix86.Code.MOV_R32_IMM32, ins.op0_register, data)
                    modified = True 
        return modified

    def _fold_constants(self, instructions: typing.List[ix86.Instruction]) -> bool:
        modified = False
        for i in range(len(instructions)):
            ins = instructions[i]
            
            if not self._is_mov_immediate(ins):
                continue 

            for j in range(i+1, len(instructions)):
                ins2 = instructions[j]

                #We have <op> reg1, imm
                if (op := self._is_arith_reg_imm(ins2, ins.op0_register)) is not None:
                    result = op(ins.immediate(1), ins2.immediate(1))
                    #print("---")
                    #print(f"Merging {ins} and {ins2}")
                    #print(f"Result is: {hex(result)}")
                    #print("---")
                    
                    instructions[i] = self._update_mov(ins, result)
                    instructions[j] = ix86.Instruction.create(ix86.Code.NOPQ)
                    modified = True
                    break
                #We have <op> reg2, reg1
                elif (op := self._is_arith_reg_reg(ins2, ins.op0_register)) is not None and self._reg_size(ins2.op1_register) < 64:
                    instructions[j] = self._update_arith(ins2, op(ins.immediate(1)))
                    modified = True
                    break
                #We have mov [mem], reg1
                elif ins2.mnemonic == ix86.Mnemonic.MOV and ins2.op0_kind == ix86.OpKind.MEMORY and ins2.op1_register == ins.op0_register:
                    mem = ix86.MemoryOperand.ctor_u64(ins2.memory_base, ins2.memory_index, ins2.memory_index_scale, ins2.memory_displacement, 8, 
                    ins2.is_broadcast, ins2.memory_segment)
                    instructions[j] = ix86.Instruction.create_mem_u32(ix86.Code.MOV_RM32_IMM32, mem, ins.immediate(1))
                    modified = True
                    break
                elif self._conflict(ins, ins2):
                    break 
        return modified

    def _remove_dead_movs(self, instructions: typing.List[ix86.Instruction]) -> bool:
        modified = False
        for i in range(len(instructions)):
            ins = instructions[i]
            
            if not self._is_mov_immediate(ins):
                continue 

            for j in range(i+1, len(instructions)):
                ins2 = instructions[j]

                if self._is_mov_immediate(ins2) and ins2.op0_register == ins.op0_register:
                    instructions[i] = ix86.Instruction.create(ix86.Code.NOPQ) 
                    modified = True 
                    break
                elif self._conflict(ins, ins2):
                    break 
        return modified

    def _merge_add_sub(self, instructions: typing.List[ix86.Instruction]) -> bool:
        modified = False
        for i in range(len(instructions)):
            ins = instructions[i]
            
            if not (ins.mnemonic in [ix86.Mnemonic.ADD, ix86.Mnemonic.SUB] and ins.op0_kind == ix86.OpKind.REGISTER and self._is_op_immediate(ins, 1)):
                continue 

            ins_is_add = ins.mnemonic == ix86.Mnemonic.ADD
            ins_imm = ins.immediate(1)
            size = self._reg_size(ins.op0_register)
            mask = (1<<size)-1

            if size > 32:
                continue

            for j in range(i+1, len(instructions)):
                ins2 = instructions[j]

                if ins2.mnemonic in [ix86.Mnemonic.ADD, ix86.Mnemonic.SUB] and ins2.op0_kind == ix86.OpKind.REGISTER and self._is_op_immediate(ins2, 1):
                    ins2_is_add = ins2.mnemonic == ix86.Mnemonic.ADD
                    ins2_imm = ins2.immediate(1)

                    sign = -1 if ins_is_add ^ ins2_is_add else 1
                    imm = (ins_imm + sign * ins2_imm) & mask

                    instructions[i] = ix86.Instruction.create_reg_u32(ix86.Code.ADD_RM32_IMM32 if ins_is_add else ix86.Code.SUB_RM32_IMM32, ins.op0_register, imm) if size <= 32 else ix86.Instruction.create_reg_u64(ins.code, ins.op0_register, imm)
                    instructions[j] = ix86.Instruction.create(ix86.Code.NOPQ) 
               
                    modified = True 
                    break
                elif self._conflict(ins, ins2):
                    break 
        return modified

    def _signed_immediate(self, imm: int, size: int) -> int:
        sb = 1 << (size-1)
        mask = (1 << size) -1
        return imm if imm & sb == 0 else -((~imm +1) & mask)

    def _remove_imul(self, instructions: typing.List[ix86.Instruction]) -> bool:
        modified = False
        for i in range(len(instructions)):
            ins = instructions[i]
            
            if ins.code not in [ix86.Code.IMUL_R16_RM16, ix86.Code.IMUL_R32_RM32, ix86.Code.IMUL_R64_RM64]:
                continue
            
            r1, r2 = ins.op0_register, ins.op1_register
            op1, op2 = None, None
            size = self._reg_size(ins.op0_register)
            mask = (1 << size) -1
            for j in range(i-1, -1, -1):
                ins2 = instructions[j]

                if self._is_mov_immediate(ins2):
                    if op1 is None and r1 == ins2.op0_register:
                        op1 = self._signed_immediate(ins2.immediate(1), size)
                    
                    if op2 is None and r2 == ins2.op0_register:
                        op2 = self._signed_immediate(ins2.immediate(1), size)

                    if op1 is not None and op2 is not None:
                        instructions[i] = self._make_mov_imm(ins.op0_register, (op1 * op2) & mask)


                #elif self._conflict(ins, ins2):
                elif self._write_to(ins2, ins.op0_register) or self._write_to(ins2, ins.op1_register):
                    break 
        return modified

    def _remove_cmp_setcc(self, instructions: typing.List[ix86.Instruction]) -> bool:
        modified = False
        for i in range(len(instructions)-1):
            ins = instructions[i]
            
            if not (ins.mnemonic == ix86.Mnemonic.CMP and ins.op0_kind == ix86.OpKind.REGISTER and self._is_op_immediate(ins, 1)) :
                continue

            insb = instructions[i+1]
            if insb.code not in [ix86.Code.SETA_RM8, ix86.Code.SETAE_RM8, ix86.Code.SETB_RM8, ix86.Code.SETBE_RM8, ix86.Code.SETE_RM8,
                ix86.Code.SETG_RM8, ix86.Code.SETGE_RM8, ix86.Code.SETL_RM8, ix86.Code.SETLE_RM8,ix86.Code.SETNE_RM8, ix86.Code.SETNO_RM8, 
                ix86.Code.SETNP_RM8, ix86.Code.SETNS_RM8, ix86.Code.SETO_RM8, ix86.Code.SETP_RM8, ix86.Code.SETS_RM8
            ]:
                continue
            
            size = self._reg_size(ins.op0_register)

            for j in range(i-1, -1, -1):
                ins2 = instructions[j]

                if self._is_mov_immediate(ins2) and ins2.op0_register == ins.op0_register:
                    rhs = ins.immediate(1) 
                    lhs = ins2.immediate(1)
                    slhs = self._signed_immediate(lhs, size)
                    srhs = self._signed_immediate(rhs, size)

                    to_set = False
                    if insb.code == ix86.Code.SETA_RM8:
                        to_set = lhs > rhs 
                    elif insb.code == ix86.Code.SETAE_RM8:
                        to_set = lhs >= rhs 
                    elif insb.code == ix86.Code.SETB_RM8:
                        to_set = lhs < rhs 
                    elif insb.code == ix86.Code.SETBE_RM8:
                        to_set = lhs > rhs 
                    
                    elif insb.code == ix86.Code.SETE_RM8:
                        to_set = lhs == rhs 
                    elif insb.code == ix86.Code.SETNE_RM8:
                        to_set = lhs != rhs 

                    elif insb.code == ix86.Code.SETG_RM8:
                        to_set = slhs > srhs 
                    elif insb.code == ix86.Code.SETGE_RM8:
                        to_set = slhs >= srhs 
                    elif insb.code == ix86.Code.SETL_RM8:
                        to_set = slhs < srhs 
                    elif insb.code == ix86.Code.SETLE_RM8:
                        to_set = slhs <= srhs 
                    else:
                        raise ValueError("CC not implemented")

                    instructions[i] = self._make_mov_imm(insb.op0_register, 1 if to_set else 0)
                    instructions[i+1] = ix86.Instruction.create(ix86.Code.NOPQ) 
                    modified = True 
                    break

                elif self._write_to(ins2, ins.op0_register):
                    break 
        return modified

    def _remove_cmovcc(self, instructions: typing.List[ix86.Instruction]) -> bool:
        modified = False
        for i in range(len(instructions)-3):
            test = instructions[i]
            mov1 = instructions[i+1]
            mov2 = instructions[i+2]
            cmov = instructions[i+3]

            if not (test.mnemonic == ix86.Mnemonic.TEST and test.op0_kind == ix86.OpKind.REGISTER and self._is_op_immediate(test, 1)) :
                continue

            if cmov.mnemonic not in [
                #ix86.Mnemonic.CMOVA, ix86.Mnemonic.CMOVAE, ix86.Mnemonic.CMOVB, ix86.Mnemonic.CMOVBE,
                #ix86.Mnemonic.CMOVG, ix86.Mnemonic.CMOVGE, ix86.Mnemonic.CMOVL, ix86.Mnemonic.CMOVLE,
                ix86.Mnemonic.CMOVE, ix86.Mnemonic.CMOVNE
            ]:
                continue

            if not self._is_mov_immediate(mov1) or not self._is_mov_immediate(mov2) or cmov.op0_register != mov2.op0_register or cmov.op1_register != mov1.op0_register:
                continue 
            
            cv1 = mov1.immediate(1)
            cv2 = mov2.immediate(1)

            size = self._reg_size(test.op0_register)

            for j in range(i-1, -1, -1):
                ins2 = instructions[j]

                if self._is_mov_immediate(ins2) and test.op0_register == ins2.op0_register:
                    lhs = ins2.immediate(1)
                    rhs = test.immediate(1)
                    slhs = self._signed_immediate(lhs, size)
                    srhs = self._signed_immediate(rhs, size)

                    to_set = False
                    """
                    if cmov.mnemonic == ix86.Mnemonic.CMOVA:
                        to_set = lhs > rhs 
                    elif cmov.mnemonic == ix86.Mnemonic.CMOVAE:
                        to_set = lhs >= rhs 
                    elif cmov.mnemonic == ix86.Mnemonic.CMOVB:
                        to_set = lhs < rhs 
                    elif cmov.mnemonic == ix86.Mnemonic.CMOVBE:
                        to_set = lhs > rhs 
                    """
                    if cmov.mnemonic == ix86.Mnemonic.CMOVE:
                        to_set = lhs & rhs  == 0
                    elif cmov.mnemonic == ix86.Mnemonic.CMOVNE:
                        to_set = lhs & rhs != 0
                    else:
                        raise ValueError("CC not implemented")

                    """
                    elif cmov.mnemonic == ix86.Mnemonic.CMOVG:
                        to_set = slhs > srhs 
                    elif cmov.mnemonic == ix86.Mnemonic.CMOVGE:
                        to_set = slhs >= srhs 
                    elif cmov.mnemonic == ix86.Mnemonic.CMOVL:
                        to_set = slhs < srhs 
                    elif cmov.mnemonic == ix86.Mnemonic.CMOVLE:
                        to_set = slhs <= srhs 
                    """
                    
                    #print(to_set)
                    instructions[i] = self._make_mov_imm(cmov.op0_register, cv1 if to_set else cv2)
                    instructions[i+1] = ix86.Instruction.create(ix86.Code.NOPQ) 
                    instructions[i+2] = ix86.Instruction.create(ix86.Code.NOPQ) 
                    instructions[i+3] = ix86.Instruction.create(ix86.Code.NOPQ) 
                    modified = True 
                    break

                elif self._write_to(ins2, test.op0_register):
                    break 
        return modified

    def is_superflous(self, i: ix86.Instruction, disp_end_va: int) -> bool:
        return self._is_mov_immediate(i) or self._is_mov_to_lvar(i, disp_end_va)

def dfs_reach_dispatcher_end(bb: BasicBlock, disptcher_end_va: int) -> typing.List[BasicBlock]:
    candidates = [bb]
    result = []
    done = set()
    while candidates:
        c = candidates.pop(0)
        if c in done:
            continue 

        done.add(c)

        if c.branch is not None:
            if c.branch.va == disptcher_end_va:
                result.append(c)
            else:
                candidates.append(c.branch)

        if c.next is not None:
            if c.next.va == disptcher_end_va:
                result.append(c)
            else:
                candidates.append(c.next)

    return result #None if not result else result[0]
        

def simplify(bb: BasicBlock, offset: int):
    cpu = CPU()
    ins = cpu.process(bb) 
    tickets = cpu.find_mov_to_local_var(offset, ins)
    bb.set_simplified(ins)
    return tickets


def detangle(start: BasicBlock, context: typing.Dict[int, BasicBlock]):
    #Step 1: delimit the dispatcher loop
    dispatcher_start = start.next
    dispatcher_end = [bb for bb in context.values() if bb.next == dispatcher_start and bb != start][0]

    #Step 1b: find the initial ticket
    initial_ticket = None 
    ticket_var_offset = None
    for i in start.instructions:
        if i.code == ix86.Code.MOV_RM32_IMM32 and i.memory_base == ix86.Register.RBP and i.memory_index == ix86.Register.NONE:
            initial_ticket = i.immediate(1)
            ticket_var_offset = i.memory_displacement
            break 

    if initial_ticket is None:
        raise ValueError("Initial ticket not found")

    #Step 2: find all the dispatching branches
    #The form is: jmp -> mov, sub, jz  (T)-> work code
    #                                 (NT)-> next dispatching branch
    tickets = {None: None}
    blocks = [dispatcher_start]
    while blocks[-1].end == BasicBlockEnd.CONDITIONAL:
        cur_block = blocks[-1]

        for i in cur_block.instructions:
            if i.code in [ix86.Code.SUB_RM32_IMM32, ix86.Code.SUB_EAX_IMM32] and i.op0_register != ix86.Register.NONE:
                tickets[i.immediate(1)] = cur_block.branch

        if len(cur_block.next.instructions) == 1 and cur_block.next.end == BasicBlockEnd.UNCONDITIONAL:
            blocks.append(cur_block.next.next)

    #Step 3: Rewire
    start.set_next(tickets[initial_ticket])
    todo = [start.next]
    done = set()
    while todo:
        block = todo.pop(0)
        if block is None or block in done:
            continue 
        done.add(block)

        for b in dfs_reach_dispatcher_end(block, dispatcher_end.va):
            tn, tb = simplify(b, ticket_var_offset)
            b.set_next(tickets[tn])
            b.set_branch(tickets[tb])
            
            todo.append(b.next)
            todo.append(b.branch)

    def ilen(i, va):
        if i.len > 0:
            return i.len 
        e = ix86.Encoder(64)
        return e.encode(i, va)

    #Step 4: layout the BBS
    va = start.va 
    code = []
    done = {}
    blocks = [(None, start)]
    while blocks:
        j, b = blocks.pop()
        if b in done:
            block_va = done[b]
            if j is not None:
                code[j] = ix86.Instruction.create_branch(code[j].code, block_va)
            else:
                code.append(ix86.Instruction.create_branch(ix86.Code.JMP_REL32_64, block_va))
                va += ilen(code[-1], va)
            continue 
        elif j is not None:
                code[j] = ix86.Instruction.create_branch(code[j].code, va)
        done[b] = va

        ni = b.new_instructions(ticket_var_offset)
        if len(ni) == 0:
            assert(b.branch is None)
        
        for i in ni:
            i.ip = va
            va += ilen(i, va)

        code.extend(ni)
        if b.branch is not None:
            blocks.append((len(code)-1, b.branch)) 

        if b.next is not None:
            blocks.append((None, b.next))
          
    encoder = ix86.BlockEncoder(64, False)
    encoder.add_many(code)
    encoded_bytes = encoder.encode(start.va)
    
    return encoded_bytes    


nt_header_offset = read_u32(0x3c)
n_sections = read_u16(nt_header_offset + 0x6)
n_sections_offset = nt_header_offset + read_u16(nt_header_offset + 0x14) + 0x18 
image_base = read_u64(nt_header_offset + 0x30)
sections = {}
for i in range(n_sections):
    section_start_offset = n_sections_offset + 0x28 * i
    
    name = pe[section_start_offset : section_start_offset + 0x8].rstrip(b"\x00").decode("iso-8859-1")
    va_size = read_u32(section_start_offset + 0x8 )
    rva = read_u32(section_start_offset + 0x0c )
    raw_size = read_u32(section_start_offset + 0x10 )
    raw_offset = read_u32(section_start_offset + 0x14 )

    sections[name] = dict(rva=rva, va_size=va_size, raw_size=raw_size, raw_offset=raw_offset)

WinMain = 0x401550
winmain_ctx = {}
winmain_bb = BasicBlock.walk_function(WinMain, winmain_ctx)


new_code = detangle(winmain_bb, winmain_ctx)

with open("FATTURA_.exe", "rb") as f:
    new_pe = bytearray(f.read())
offset = va_to_offset(WinMain)
for i in range(len(new_code)):
    new_pe[offset+i] = new_code[i]
with open("TEST.exe", "wb") as f:
    f.write(new_pe)    